import random
import argparse
import sys, string, glob, re, os
import numpy as np
import time,copy

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import TensorDataset,DataLoader
from torch.nn import functional as F

from model_meta_cate import MetaCate,MultitaskCate,MAMLCate,SimpleCate,MAMLCateIPM
from econml.metalearners import SLearner,XLearner,TLearner
from sklearn.ensemble import GradientBoostingRegressor,GradientBoostingClassifier
from econml.dml import CausalForestDML
from econml.dr import LinearDRLearner
from sklearn.linear_model import LassoCV
from sklearn.linear_model import Ridge, LogisticRegression, LinearRegression
        
def traineval_epoch(dataloader,model,optimizer,args,is_train):
    if model is not None:
        if is_train == 1:
            model.train()
        else:
            model.eval()

    retain_graph = False
    if args.method == 5 or args.method == 6 or args.method == 8 or args.method == 9: #MAML
        retain_graph = True
        

    total_loss = 0.0
    total_err = 0.0
    total_cnt = 0
    for batch_idx,(x,y,a,tau,tauhat,task) in enumerate(dataloader):
        n_task,n_sample,n_att = x.shape
        x_spt = torch.zeros_like(x[:,:args.n_spt0+args.n_spt1,:])
        y_spt = torch.zeros_like(y[:,:args.n_spt0+args.n_spt1,:])
        a_spt = torch.zeros_like(a[:,:args.n_spt0+args.n_spt1,:])
        x_qry = torch.zeros_like(x[:,:args.n_qry*2,:])
        y_qry = torch.zeros_like(y[:,:args.n_qry*2,:])
        a_qry = torch.zeros_like(a[:,:args.n_qry*2,:])
        tauhat_qry = torch.zeros_like(tauhat[:,:args.n_qry*2,:])
        tau_qry = torch.zeros_like(tau[:,:args.n_qry*2,:])        
        for i in range(n_task):
            a1idx = (a[i,:,:]==1).squeeze(-1)
            a0idx = (a[i,:,:]==0).squeeze(-1)
            x0 = x[i,a0idx,:]
            y0 = y[i,a0idx,:]
            tauhat0 = tauhat[i,a0idx,:]
            tau0 = tau[i,a0idx,:]
            x1 = x[i,a1idx,:]
            y1 = y[i,a1idx,:]
            tauhat1 = tauhat[i,a1idx,:]
            tau1 = tau[i,a1idx,:]
            if is_train == 1:
                perm0 = np.random.permutation(x0.shape[0])
                perm1 = np.random.permutation(x1.shape[0])
            else:
                perm0 = np.arange(x0.shape[0])
                perm1 = np.arange(x1.shape[0])
            x_qry[i,:args.n_qry,:] = x0[perm0[:args.n_qry],:]
            x_qry[i,args.n_qry:,:] = x1[perm1[:args.n_qry],:]
            x_spt[i,:args.n_spt0,:] = x0[perm0[args.n_qry:args.n_qry+args.n_spt0],:]
            x_spt[i,args.n_spt0:,:] = x1[perm1[args.n_qry:args.n_qry+args.n_spt1],:]

            y_qry[i,:args.n_qry,:] = y0[perm0[:args.n_qry],:]
            y_qry[i,args.n_qry:,:] = y1[perm1[:args.n_qry],:]
            y_spt[i,:args.n_spt0,:] = y0[perm0[args.n_qry:args.n_qry+args.n_spt0],:]
            y_spt[i,args.n_spt0:,:] = y1[perm1[args.n_qry:args.n_qry+args.n_spt1],:]

            tau_qry[i,:args.n_qry,:] = tau0[perm0[:args.n_qry],:]
            tau_qry[i,args.n_qry:,:] = tau1[perm1[:args.n_qry],:]
            tauhat_qry[i,:args.n_qry,:] = tauhat0[perm0[:args.n_qry],:]
            tauhat_qry[i,args.n_qry:,:] = tauhat1[perm1[:args.n_qry],:]
            
        a_spt[:,args.n_spt0:,:] = torch.ones_like(a_spt[:,args.n_spt0:,:])
        a_qry[:,args.n_qry:,:] = torch.ones_like(a_qry[:,args.n_qry:,:])
        if args.method == 4:
            loss,err = model.forward_eval(x_spt,y_spt,a_spt,x_qry,y_qry,a_qry,tauhat_qry,tau_qry,args.n_spt0,args.n_spt1,args.n_qry)
        elif args.method >= 10: #causal forest
            x_spt = x_spt.numpy()
            a_spt = a_spt.numpy()==1
            y_spt = y_spt.numpy()
            x_qry = x_qry.numpy()
            tauhat_qry = tauhat_qry.numpy()
            tau_qry = tau_qry.numpy()
            loss = 0.0
            err = 0.0
            for i in range(n_task):
                if args.method == 10: #causal forest
                    est = CausalForestDML(model_t=LassoCV(cv=2), model_y=LassoCV(cv=2))
                elif args.method == 11: #xlearner
                    est = XLearner(models=GradientBoostingRegressor(),
                                   propensity_model=GradientBoostingClassifier(),
                                   cate_models=GradientBoostingRegressor())
                elif args.method == 12: #slearner
                    est = SLearner(overall_model=GradientBoostingRegressor())
                elif args.method == 13: #tlearner
                    est = TLearner(models=GradientBoostingRegressor())
                elif args.method == 14: #drlearner
                    est = LinearDRLearner(model_propensity=GradientBoostingClassifier(),
                                          model_regression=GradientBoostingRegressor())
                elif args.method == 15 or args.method == 19: #xlearner-linear
                    est = XLearner(models=LinearRegression(),
                                   propensity_model=LogisticRegression(),
                                   cate_models=LinearRegression())
                elif args.method == 16 or args.method == 20: #slearner-linear
                    #est = SLearner(overall_model=LinearRegression())
                    est = SLearner(overall_model=Ridge())
                elif args.method == 17 or args.method == 21: #tlearner-linear
                    est = TLearner(models=LinearRegression())
                elif args.method == 18 or args.method == 22: #drlearner-linear
                    est = LinearDRLearner(model_propensity=LogisticRegression(),
                                          model_regression=LinearRegression())

                est.fit(np.ravel(y_spt[i,:,:]),np.ravel(a_spt[i,:,:]),X=x_spt[i,:,:])
                if args.method >= 19 and args.method <= 22:
                    tauhathat_spt = est.effect(x_spt[i,:,:])
                    tauhathat_qry = tauhathat_spt.mean()
                else:
                    tauhathat_qry = est.effect(x_qry[i,:,:])
                loss += np.mean((tauhathat_qry-tauhat_qry[i,:,:])**2)
                err += np.mean((tauhathat_qry-tau_qry[i,:,:])**2)
        else:
            loss,err = model(x_spt,y_spt,a_spt,x_qry,y_qry,a_qry,tauhat_qry,tau_qry,args.n_spt0,args.n_spt1,args.n_qry)            
        
        if is_train == 1:
            optimizer.zero_grad()
            loss.backward(retain_graph=retain_graph)
            optimizer.step()
        total_cnt += n_task
        if args.method < 10:
            total_loss += loss.detach().item()*n_task
            total_err += err.detach().item()*n_task
        else:
            total_loss += loss
            total_err += err
            
    return total_loss/total_cnt,total_err/total_cnt

def multitask_traineval_epoch(dataloader,model,optimizer,args,is_train):
    if is_train == 1:
        model.train()
    else:
        model.eval()

    total_loss = 0.0
    total_err = 0.0
    total_cnt = 0
    for batch_idx,(x,y,a,tau,tauhat,task) in enumerate(dataloader):
        n_task = x.shape[0]
        loss,err = model(x,y,a,tau,task)
        if is_train == 1:
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        total_loss += loss.detach().item()*n_task
        total_err += err.detach().item()*n_task
        total_cnt += n_task
    return total_loss/total_cnt,total_err/total_cnt

def train(trainloader,validloader,testloader,model,optimizer,args):
    valid_errs = []
    bestmodel = copy.deepcopy(model)

    for epoch_i in range(args.epoch):
        if args.method != 4:
            train_loss,train_err = traineval_epoch(trainloader,model,optimizer,args,is_train=1)
            valid_loss,valid_err = traineval_epoch(validloader,model,optimizer,args,is_train=0)
            test_loss,test_err = traineval_epoch(testloader,model,optimizer,args,is_train=0)
        else:
            train_loss,train_err = multitask_traineval_epoch(trainloader,model,optimizer,args,is_train=1)
            valid_loss,valid_err = traineval_epoch(validloader,model,optimizer,args,is_train=0)
            test_loss,test_err = traineval_epoch(testloader,model,optimizer,args,is_train=0)            
        print('%d train:loss: %f err: %f valid:loss: %f err: %f test:loss: %f err: %f'%(epoch_i,train_loss,train_err,valid_loss,valid_err,test_loss,test_err))
                                                     
        valid_errs += [valid_err]
        
        if valid_err <= min(valid_errs):
            bestmodel = copy.deepcopy(model)
            
    if args.out:
        model_state_dict = bestmodel.state_dict()
        checkpoint = {'model': model_state_dict,'settings': args}
        torch.save(checkpoint,"%s.chkpt"%args.out)

    return bestmodel

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--fn',type=str,default="",help='data file, e.g., data/csv/PenDigits_withoutdupl_norm_v01_t100/I10/fns0.npz')
    parser.add_argument('--epoch','-e',type=int,default=500,help='epochs')
    parser.add_argument('--n_hiddens',type=str,default='32,32',help='#hidden units')

    parser.add_argument('--n_spt0',type=int,default=3,help='#support sample for class 0')
    parser.add_argument('--n_spt1',type=int,default=3,help='#support sample for class 1')
    parser.add_argument('--n_qry',type=int,default=20,help='#query sample for each class')
    parser.add_argument('--n_qry_train',type=int,default=None,help='#query sample for each class for training')
    parser.add_argument('--r_train',type=float,default=0.7,help='ratio of training tasks')
    parser.add_argument('--r_valid',type=float,default=0.1,help='ratio of validation tasks')

    parser.add_argument('--weight_decay',type=float,default=1e-5)
    parser.add_argument('--dropout', type=float, default=0.0)
    parser.add_argument('--method',type=int,default=0,help='0:proposed 1:a,y,ytilde-loss')
    parser.add_argument('--second_method',type=int,default=2,help='0:RA 1:PA 2:DR 3:plugin')    
    parser.add_argument('--seed',type=int,default=0,help='random seed')
    parser.add_argument('--local_epoch',type=int,default=5,help='MAML inner epoch')
    parser.add_argument('--local_lr',type=float,default=1e-3,help='MAML inner lr')
    parser.add_argument('--batch_size', type=int, default=32)
    parser.add_argument('--no_cuda', action='store_true')
    parser.add_argument('--lr', type=float, default=1e-3)
    parser.add_argument('--initmodel', default=None)
    parser.add_argument('--out', default=None)
    parser.add_argument('--use_encoder',type=int,default=1,help='use support set for neural net (1) or not (0)')
    parser.add_argument('--n_dataset',type=int,default=0,help='number of tasks for training')
    parser.add_argument('--n_sample',type=int,default=0,help='number of samples per task')
    parser.add_argument('--regression',type=int,default=0,help='0:classification 1:regression')

    parser.add_argument('--GP',type=int,default=0,help='0:linear 1:GP')
    parser.add_argument('--treatment_as_feature',type=int,default=0,help='0:separamte mu 1:mu with treatment as feature')
    parser.add_argument('--encoder',type=int,default=1,help='0:no-encoder 1:neural net')
    parser.add_argument('--use_truetau',type=int,default=0,help='0:use estimate tau, 1:use true tau')
    parser.add_argument('--estimate_weight',type=int,default=0,help='0:no 1:estimate IPM loss weight')
                        
    args = parser.parse_args()    
    args.cuda = not args.no_cuda and torch.cuda.is_available()
    device = torch.device("cuda" if not args.no_cuda and torch.cuda.is_available() else "cpu")
    args.device = device
    np.random.seed(args.seed)
    random.seed(args.seed)
    torch.manual_seed(args.seed)

    print(args)

    dat = np.load(args.fn)
    x = dat['x']
    y = dat['y']
    a = dat['a']
    tau = dat['tau']
    tauidx = 3
    #tauhat = dat['tauhat'][:,:,[tauidx]]
    if args.use_truetau == 1:
        tauhat = dat['tau']
    else:
        tauhat = dat['tauhat']
    n_task,n_sample,n_att = x.shape
    n_train = int(args.r_train*n_task)
    n_valid = int(args.r_valid*n_task)
    n_test = n_task-n_train-n_valid
    if args.n_dataset > 0 and n_train > args.n_dataset:
        n_train = args.n_dataset

    if args.n_sample > 0 and n_sample > args.n_sample:
        n_sample = args.n_sample
        
    x_test = torch.FloatTensor(x[:n_test:,:n_sample,:]).to(args.device)
    x_valid = torch.FloatTensor(x[n_test:n_test+n_valid,:n_sample,:]).to(args.device)
    x_train = torch.FloatTensor(x[n_test+n_valid:n_test+n_valid+n_train,:n_sample,:]).to(args.device)
    y_test = torch.FloatTensor(y[:n_test:,:n_sample,:]).to(args.device)
    y_valid = torch.FloatTensor(y[n_test:n_test+n_valid,:n_sample,:]).to(args.device)
    y_train = torch.FloatTensor(y[n_test+n_valid:n_test+n_valid+n_train,:n_sample,:]).to(args.device)
    a_test = torch.FloatTensor(a[:n_test:,:n_sample,:]).to(args.device)
    a_valid = torch.FloatTensor(a[n_test:n_test+n_valid,:n_sample,:]).to(args.device)
    a_train = torch.FloatTensor(a[n_test+n_valid:n_test+n_valid+n_train,:n_sample,:]).to(args.device)
    tau_test = torch.FloatTensor(tau[:n_test:,:n_sample,:]).to(args.device)
    tau_valid = torch.FloatTensor(tau[n_test:n_test+n_valid,:n_sample,:]).to(args.device)
    tau_train = torch.FloatTensor(tau[n_test+n_valid:n_test+n_valid+n_train,:n_sample,:]).to(args.device)
    tauhat_test = torch.FloatTensor(tauhat[:n_test:,:n_sample,:]).to(args.device)
    tauhat_valid = torch.FloatTensor(tauhat[n_test:n_test+n_valid,:n_sample,:]).to(args.device)
    tauhat_train = torch.FloatTensor(tauhat[n_test+n_valid:n_test+n_valid+n_train,:n_sample,:]).to(args.device)

    args.n_hiddens = list(map(int,args.n_hiddens.split(',')))

    task_train = torch.arange(n_train).to(torch.long).to(args.device)
    task_valid = torch.arange(n_valid).to(torch.long).to(args.device)
    task_test = torch.arange(n_test).to(torch.long).to(args.device)
    
    traindataset = TensorDataset(x_train,y_train,a_train,tau_train,tauhat_train,task_train)
    validdataset = TensorDataset(x_valid,y_valid,a_valid,tau_valid,tauhat_valid,task_valid)
    testdataset = TensorDataset(x_test,y_test,a_test,tau_test,tauhat_test,task_test)
    trainloader = DataLoader(traindataset,batch_size=args.batch_size,shuffle=True)
    validloader = DataLoader(validdataset,batch_size=args.batch_size,shuffle=False)
    testloader = DataLoader(testdataset,batch_size=args.batch_size,shuffle=False)

    print("#train: %d #valid: %d #test: %d #sample: %d #att: %d"%(n_train,n_valid,n_test,n_sample,n_att))

    start = time.time()
    if args.method == 4:
        model = MultitaskCate(n_task=n_train,n_att=n_att,n_hiddens=args.n_hiddens,dropout=args.dropout,
                              second_method=args.second_method)
    elif args.method == 5 or args.method == 6:
        model = MAMLCate(n_att=n_att,n_hiddens=args.n_hiddens,dropout=args.dropout,
                         second_method=args.second_method,method=args.method,local_lr=args.local_lr,local_epoch=args.local_epoch)
    elif args.method == 7:
        model = SimpleCate()
        args.epoch = 0
    elif args.method == 8 or args.method == 9:
        model = MAMLCateIPM(n_att=n_att,n_hiddens=args.n_hiddens,dropout=args.dropout,
                            method=args.method,local_lr=args.local_lr,local_epoch=args.local_epoch,estimate_weight=args.estimate_weight)
    elif args.method >= 10: #causal effect
        model = None
        args.epoch = 0
    else:
        model = MetaCate(n_att=n_att,n_hiddens=args.n_hiddens,dropout=args.dropout,
                         method=args.method,second_method=args.second_method,GP=args.GP,encoder=args.encoder,treatment_as_feature=args.treatment_as_feature)
    if model is not None:
        model.to(args.device)
        if args.initmodel:
            model.load_state_dict(torch.load("%s.chkpt"%args.initmodel)['model'],strict=False)

    if args.epoch > 0:
        optimizer = torch.optim.Adam(model.parameters(),lr=args.lr,weight_decay=args.weight_decay)    
        model = train(trainloader,validloader,testloader,model,optimizer,args)
    else:
        optimizer = None

    train_time = (time.time()-start)
    start = time.time()
    test_loss,test_err = traineval_epoch(testloader,model,optimizer,args,is_train=0)
    test_time = (time.time()-start)

    print("train:time: %f test:time: %f loss: %f err: %f"%(train_time,test_time,test_loss,test_err))
